from typing import Container as Includer

from functools import partial

from swimport.__util__ import *
from swimport.model import Container, NameTrigger, SwimRule, object_swimporting, Source, SwimportingsResult, \
    RawSource, Function
from swimport.swim import Swim
from swimport.typeswim import TypeSwimporting, BuiltinTypemap, FunctionBody


class ContainerNameTrigger(Container.Trigger, NameTrigger):
    """A trigger for container names"""


class ContainerKindTrigger(Container.Trigger):
    """A trigger that filters containers by kind"""
    all_kinds = frozenset((Container.Kind.Union,
                           Container.Kind.Class,
                           Container.Kind.Struct))

    def __init__(self, valid_kind: Includer[Container.Kind]):
        """
        :param valid_kind: valid kinds of containers to accept
        """
        self.valid_kinds = valid_kind

    def is_valid(self, rule: SwimRule, obj: Container) -> bool:
        return super().is_valid(rule, obj) \
               and obj.kind in self.valid_kinds


@Container.set_default_trigger
def _(arg):
    """
    default trigger, trigger depends on arg's type:
    str: name trigger
    Container.Kind: only accept containers of that kind
    Collection: only accept containers of kinds inside the collection
    """
    if isinstance(arg, str):
        return ContainerNameTrigger(arg)
    if isinstance(arg, Container.Kind):
        return ContainerKindTrigger((arg,))
    if isinstance(arg, Includer):
        return ContainerKindTrigger(arg)
    raise TypeError('cannot parse argument as container trigger')


def container_type_import(name, args):
    p_port = partial(TypeSwimporting, name + '*',
                     to_py=BuiltinTypemap(name + '*'),
                     to_cpp=BuiltinTypemap(name + '*'),
                     to_cpp_check=BuiltinTypemap(name + '*'),
                     to_cpp_check_precedence=...)(**args)
    yield p_port
    p_to_py_func = p_port.to_py_func
    if p_to_py_func:
        to_py = FunctionBody.combine(p_to_py_func)(f"""     
                                                    return {p_to_py_func.function_name}(({name}*)(&input), ok, PACKED_UD(0));
                                                    """)
    else:
        to_py = None
    p_to_cpp_func = p_port.to_cpp_func
    if p_to_cpp_func:
        to_cpp = FunctionBody.combine(p_to_cpp_func)(f"""
                                                    auto ptr = {p_to_cpp_func.function_name}(input, ok, PACKED_UD(0));
                                                    if (!ok || PyErr_Occurred() != nullptr)
                                                        SWIM_RAISE_UNSPECIFIED;
                                                    return *ptr;
                                                    """)
    else:
        to_cpp = None
    p_to_cpp_check_func = p_port.to_cpp_check_func
    if p_to_cpp_check_func:
        to_cpp_check = FunctionBody.combine(p_to_cpp_check_func)(f"""
                                                                    return {p_to_cpp_check_func.function_name}(input, PACKED_UD(0));
                                                                    """)
    else:
        to_cpp_check = None

    port = partial(TypeSwimporting, name,
                   to_py=to_py,
                   to_cpp=to_cpp,
                   to_cpp_check=to_cpp_check,
                   to_cpp_check_precedence=...)(**args)
    yield port


@Container.set_default_behaviour
class ContainerBehaviour(Container.Behaviour):
    def __init__(self, out=True, inline=False, **type_args):
        """
        :param out: whether to prepare output typemaps for the type
        :param inline: whether to scope the container body in an inline scope instead of raw scope
        :param iter_maps: whether to apply iterator typemaps
        """
        self.out = out
        self.inline = inline
        self.type_args = type_args

    @object_swimporting
    def wrap(self, rule: SwimRule, obj: Container, swim):
        body_scope = Swim.add_inline if self.inline else Swim.add_raw

        swim.add_comment(obj)
        body = [obj.kind.value + ' ' + obj.name + '{', '\tpublic:']
        for m in obj.members:
            if m.access != 'public':
                raise ValueError('Default Container.Behaviour can only handle simple aggregate containers.'
                                 'Use the ContainerSwim object')
            body.append('\t\t' + m.body + ';')
        body.append('};')
        body_scope(swim, body)
        swim(container_type_import(obj.name, self.type_args))


class ContainerSwim:
    """A swimporting aggregate handling the scope of a single container."""

    # todo ability to mark methods as @abstractmethods

    def __init__(self, trigger, source: Source, *, director=False, wrapper_superclass=None, **type_args):
        """
        :param trigger: the trigger (or default trigger argument) with which to identify the proper container to import
        :param source: the source object to get the container from
        :param director: whether to insert a director for this class, allowing it to be subclassed from python
        :param type_args: additional arguments to the internal typeswimporting
        """
        self.director = director
        if self.director:
            type_args.setdefault('to_py', None)

        self.type_args = type_args
        self.wrapper_superclass = wrapper_superclass

        trigger = Container.Trigger(trigger)
        qualified = None
        for c in source:
            if trigger.is_valid(None, c):
                if qualified is None:
                    qualified = c
                else:
                    raise Exception(f'multiple containers matched: {c}, {qualified}, ...')
        if not qualified:
            raise Exception('no container matched')
        self.container: Container = qualified

        bases = list(self.container.public_bases)
        if bases:
            bases_str = ': public ' + ', public '.join(bases)
        else:
            bases_str = ''

        self._in_body = [
            Swim.add_raw(f'{self.container.kind.value} {self.container.name}{bases_str} {{')
        ]
        for m in self.container.members:
            if m.is_static:
                continue
            self._in_body.append(Swim.add_raw(f'{m.access}: {m.body}'))
        self._in_body.append(Swim.add_raw('public:'))

        self.extend = []

        self.post_body = []

    def __call__(self, rule: SwimRule) -> SwimportingsResult:
        """Process a rule over all the public members and functions of the container"""
        ret = SwimportingsResult()
        for o in self._importable_cpp_objects():
            if o.access != 'public':
                continue
            porting = rule.do(o)
            if porting is not None:
                self._in_body.append(porting)
                obj = getattr(porting, 'object', None)
                if obj:
                    ret += obj.name
                else:
                    ret += 1
        return ret

    def __iter__(self):
        yield from container_type_import(self.container.name, self.type_args)
        if self.director:
            yield Swim.add_raw(f'%feature("director") {self.container.name};')
        if self.wrapper_superclass:
            yield Swim.add_raw(f'%feature("python:abc",{self.wrapper_superclass}) {self.container.name};')
        yield from self._in_body
        if self.extend:
            yield Swim.add_raw('%extend {')
            yield from self.extend
            yield Swim.add_raw('}')
        yield Swim.add_raw('};')
        yield from self.post_body

    def _importable_cpp_objects(self):
        yield from (x for x in self.container.members if x.is_static)
        yield from self.container.methods

    def extend_method(self, sign, body, behaviour=...):
        """
        Add an extension method to the container
        :param sign: the signature of the method.
        :param body: the (cpp) body of the method. use $self instead of this.
        :param behaviour: the behaviour with which to import the method
        """
        if behaviour is ...:
            behaviour = Function.Behaviour()

        behaviour = behaviour.replace(body=body, body_scope=Swim.add_raw)

        i = iter(RawSource(sign))
        function = next(i, None)
        if next(i, None):
            raise Exception('multiple signatures detected')
        if not function:
            raise Exception('no signature')

        self.extend.append(behaviour.do(function))

    def extend_py_def(self, symbol, params, body, wrapper: str = ...):
        """
        Extend the container with a python method. The method is only added after the class is created.
        :param symbol: the name of the function to add/replace
        :param params: the parameters of the function
        :param body: the (python) body of the function
        :param wrapper: if present, will wrap the new function (e.g. "staticmethod", "property").
            Default will copy the type of the previous value, if any.
        """

        py_code = ['__temp_store = getattr(' + self.container.name + ',"""' + symbol + '""",None)',
                   f'def __temp_def({params}):']
        py_code.extend(scope_lines(body, '', '', inline=False))
        if wrapper is ...:
            py_code.extend(['if isinstance(__temp_store, (classmethod, staticmethod, property)):',
                            '\t__temp_def = type(__temp_store)(__temp_def)'])
        elif wrapper:
            py_code.append(f'__temp_def = {wrapper}(__temp_def)')
        py_code.extend(['__temp_def.prev = __temp_store', self.container.name + '.' + symbol + ' = __temp_def',
                        'del __temp_store, __temp_def'])
        self.post_body.append(Swim.add_python(py_code))
        self._in_body.append(Swim.add_python(
            [
                f"if '''{symbol}''' not in locals():",
                ("\t@" + wrapper) if wrapper and wrapper is not ... else "",
                f"\tdef {symbol}({params}): pass"
            ]
        ))



from typing import List, Optional, Union, TextIO, MutableMapping, Any, Tuple, Type

from functools import lru_cache
from os import PathLike, fspath
import itertools as it

import swimport.pools as pools_pkg
from swimport.model import SwimportAggregate, CPPObject, Swimporting, Function
from swimport.typeswim import TypeSwimporting
from swimport.__util__ import *


class Delaying:
    """A wrapper descriptor for functions so they can be called as parting functions from the owner class"""

    def __init__(self, func):
        self.__func__ = func

    @lru_cache(None)
    def parting(self, owner_cls):
        def ret(*args, **kwargs):
            if args and isinstance(args[0], owner_cls):
                return self.__func__(*args, **kwargs)
            return lambda s: self.__func__(s, *args, **kwargs)

        return ret

    def __get__(self, instance, owner):
        if instance is None:
            return self.parting(owner)
        return self.__func__.__get__(instance)

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)


class Swim(SwimportAggregate):
    """The main swim object, absorbs all the lines and scopes given by swimportings"""

    def __init__(self, module_name: Optional[str], disclose=True, verbosity: Verbosity = Verbosity.info, **kwargs):
        """
        :param module_name: the name of the module, if this is set, the module pool will be automatically added
        :param disclose: whether to automatically add the disclose pool
        :param verbosity: the verbosity of the swim
        :param kwargs: additional arguments fed to the module pool
        """
        # note: special swimportings can place their markers here even though they are NOT cppObjects
        self.imported: MutableMapping[CPPObject, Any] = {}
        self.types: MutableMapping[str, TypeSwimporting] = {}
        self.lines: List[str] = []
        self.module_name = module_name
        self.id_mark = 'SWIM_MAIN' if not module_name else ('SWIM_' + module_name.upper())
        self.verbosity = verbosity

        if disclose:
            self(pools_pkg.pools.disclose)
        if module_name:
            self(pools_pkg.pools.module(module_name, **kwargs))

    def apply_porting(self, swimporting: Swimporting) -> Union[str, int]:
        """
        Apply a single porting to the swim
        :returns: whether or not the porting succeeded
        """
        if swimporting(self):
            obj = getattr(swimporting, 'object', None)
            if obj:
                return obj.name
            return 1
        return 0

    def _add_scope(self, *args, **kwargs):
        """
        adds lines to the swim object. with the arguments forwarded to __util__.scope_lines
        :returns int: the number of lines added
        """
        lines = scope_lines(*args, **kwargs)
        self.lines.extend(lines)
        return len(lines)

    @Delaying
    def add_raw(self, lines):
        """
        add lines to the swim without an enclosing scope
        """
        return self._add_scope(lines, None, None, indent=None, inline=False)

    @Delaying
    def add_extend(self, lines, name=''):
        """
        add lines to the swim without an enclosing scope
        """
        return self._add_scope(lines, '%extend ' + name + '{', '}', indent=None, inline=False)

    @Delaying
    def add_insert(self, lines):
        """
        add lines to the swim with an insert scope %{ %}
        """
        return self._add_scope(lines, '%{', '%}')

    @Delaying
    def add_init(self, lines):
        """
        add lines to the swim with an %init scope
        """
        return self._add_scope(lines, '%init %{', '%}')

    @Delaying
    def add_begin(self, lines):
        """
        add lines to the swim with an %init scope
        """
        return self._add_scope(lines, '%begin %{', '%}')

    @Delaying
    def add_inline(self, lines):
        """
        add lines to the swim with an %inline scope
        """
        return self._add_scope(lines, '%inline %{', '%}')

    @Delaying
    def add_python(self, lines):
        """
        add lines to the swim with a %pythoncode scope
        """
        return self._add_scope(lines, '%pythoncode %{', '%}', inline=False)

    @Delaying
    def add_python_begin(self, lines):
        """
        add lines to the swim with a %pythonbegin scope
        """
        return self._add_scope(lines, '%pythonbegin %{', '%}', inline=False)

    @Delaying
    def add_feature(self, lines, pattern, keyword):
        if lines:
            return self._add_scope(lines, '%' + keyword + ' ' + pattern + ' %{', '%}', inline=False)
        else:
            return self.add_raw('%' + keyword + ' ' + pattern + ';')

    @Delaying
    def add_python_append(self, method: Function, lines):
        """
        add lines to the swim with a %pythonappend scope. requires the method to which to bind.
        """
        return self.add_feature(lines, method.pattern, 'pythonappend')

    @Delaying
    def add_python_prepend(self, method: Function, lines):
        """
        add lines to the swim with a %pythonprepend scope. requires the method to which to bind.
        """
        return self.add_feature(lines, method.pattern, 'pythonprepend')

    @Delaying
    def add_exception_check(self, lines):
        """
        add lines to the swim with a %exception scope. requires the method to which to bind.
        """
        return self.add_feature(lines, '', 'exception')

    @Delaying
    def add_contract(self, method: Function, lines):
        """
        add lines to the swim with a %contract scope. requires the method to which to bind.
        """
        return self.add_feature(lines, method.pattern, 'contract')

    @Delaying
    def add_comment(self, lines, verbosity_level=Verbosity.info):
        """
        adds lines as a collection of singleline comments
        """
        if self.verbosity < verbosity_level:
            return 0
        return self._add_scope(lines, None, None, '// ', inline=False)

    @Delaying
    def add_comment_ml(self, lines, verbosity_level=Verbosity.info):
        """
        adds lines as a multiline comment
        """
        if self.verbosity < verbosity_level:
            return 0
        return self._add_scope(lines, '/*', '*/', '* ', inline=False)

    def create_unique_id(self, id_seed: str):
        ret = f'__SWIMPORT_{self.id_mark}_{clean_identifier(id_seed)}'
        return ret

    def add_function(self, return_type: str, function_name: str, lines, arguments: list, add_scope=add_raw, *,
                     name_abs=False, fragments=()) -> str:
        """
        add a cpp function to the swim file
        :param return_type: the return type of the function
        :param function_name: the name of the function. Note: the name will be mutated to ensure no collisions,\
        the true name of the function will be returned.
        :param lines: the code lines of the function
        :param arguments: a list of cpp parameter identifiers for the function
        :param add_scope: the scope of the function deceleration.
        :param name_abs: if set, regular name mangling will be ignored the function's identifier will be identical
            to function_name. note that this identifier will not be stored in the swim's translated_ids.
        :return: the true identifier of the function.
        """
        if name_abs:
            name = function_name
        else:
            name = self.create_unique_id(function_name)
        frag_str = (',fragment="' + ','.join(fragments) + '"') if fragments else ''
        lines = clean_source(lines)
        lines = [
            '%fragment ("' + name + '","header"' + frag_str + '){',
            return_type + ' ' + name + '(' + ', '.join(arguments) + ') {',
            *('\t' + l for l in lines),
            '}',
            '}'
        ]
        add_scope(self, lines)
        return name

    def add_nl(self, n=1):
        """
        add a blank line. Note: the blank line will be ignored if the swim's verbosity is at Verbosity.Critical or lower
        :param n: the number of blank lines to add.
        :return:  the nuber of lines added.
        """
        if self.verbosity <= Verbosity.critical:
            return 0
        if n == 1:
            self.lines.append('')
        else:
            self.lines.extend('' for _ in range(n))
        return n

    def __str__(self):
        if self.module_name:
            return f'swim for module {self.module_name}'
        return super().__str__()

    def __iter__(self):
        """
        Get an iterable of the lines stored in the swimport.
        """
        return (yield from self.lines)

    def write(self, dst: Union[str, PathLike, TextIO]):
        """
        write the lines of the swim to a file or path
        """
        if isinstance(dst, (str, PathLike)):
            dst = open(fspath(dst), 'w')
            own = True
        else:
            own = False

        dst.writelines(
            it.chain.from_iterable(zip(self, it.repeat('\n')))
        )

        if own:
            dst.close()

    @staticmethod
    def _normalize_to_cpp_type(t)->str:
        if isinstance(t, str):
            return t
        type_dict = {
            int: 'int',
            float: 'double',
            bool: 'bool',
            str: 'std::string',
        }
        ret = type_dict.get(t)
        if ret:
            return ret
        if isinstance(t, tuple):
            return 'std::tuple<' + ', '.join(Swim._normalize_to_cpp_type(i) for i in t) + '>'
        if isinstance(t, list) and len(t) == 1:
            return 'std::vector<' + Swim._normalize_to_cpp_type(t[0]) + '>'
        if isinstance(t, set) and len(t) == 1:
            return 'std::unordered_set<' + Swim._normalize_to_cpp_type(next(iter(t))) + '>'
        if isinstance(t, dict) and len(t) == 1:
            k, v = next(iter(t.items()))
            return 'std::unordered_map<' + Swim._normalize_to_cpp_type(k) + ', ' + Swim._normalize_to_cpp_type(v) + '>'
        raise ValueError(f"can't parse as c++ type: {t}")

    def get_porting(self, type_: Union[str, TypeSwimporting, Tuple[str, TypeSwimporting]]) -> Tuple[str, TypeSwimporting]:
        if isinstance(type_, tuple) and len(type_) == 2\
                and isinstance(type_[0], str) and isinstance(type_[1], TypeSwimporting):
            return type_
        if isinstance(type_, TypeSwimporting):
            return type_.cpp_name, type_
        try:
            type_ = self._normalize_to_cpp_type(type_)
            return type_, self.types[type_]
        except KeyError as e:
            raise LookupError(f'type {type_} has not been swimported (did you remember to use pools.primitive?)') from e



from __future__ import annotations
from typing import Iterable, Optional, Union, TypeVar, Sequence

import re
from enum import IntEnum
from functools import lru_cache, partial, update_wrapper
from string import Formatter
from textwrap import dedent
from itertools import chain
from inspect import Parameter, signature

T = TypeVar('T')


class Verbosity(IntEnum):
    """verbosity of the output file"""
    none = 0
    critical = 1
    info = 2
    debug = 3


def clean_source(source) -> Sequence[str]:
    """
    converts string-like objects to a sequence of strings. includes stripping, and dedenting
    >>> clean_source("\\n\\ta\\n\\tb\\n\\tc")
    ['a', 'b', 'c']
    >>> clean_source("\\ta\\n\\tb\\n\\tc\\n")
    ['a', 'b', 'c']
    """
    if isinstance(source, str):
        return dedent(source).strip().splitlines(keepends=False)
    if isinstance(source, Sequence):
        return source
    if isinstance(source, Iterable):
        return list(source)
    if type(source).__str__ is not object.__str__:
        return clean_source(str(source))
    raise TypeError(f'object {source} cannot be interpreted as string or lines')


def scope_lines(lines, start: Optional[str], end: Optional[str], indent: Optional[str] = '\t',
                inline: Union[bool, type(...)] = ..., inline_sep: Optional[str] = None):
    """
    wraps lines in a scope, and returns the lines as a list
    :param lines: the lines (or string-like, ran through clean_source)
    :param start: the opener of the scope
    :param end: the closer of the scope
    :param indent: the indentation to apply to each line of the scope
    :param inline: whether to inline the scope. or ... to only inline if the source is in a single line.
    :param inline_sep: the separator to use when joining multiple lines in an inlined scope.
    :return: a list of the scope lines
    """
    lines = clean_source(lines)
    if not lines:
        return []

    if inline is ...:
        inline = len(lines) == 1

    if inline:
        if len(lines) == 1:
            s = lines[0]
        else:
            if inline_sep is None:
                raise ValueError('seperator must be entered for multiline, inlined scope')
            s = inline_sep.join(lines)
        start = '' if not start else start + ' '
        end = '' if not end else ' ' + end
        return [start + s + end]

    if not (start or end or indent):
        return lines

    ret = []

    if start:
        ret.append(start)

    if indent:
        lines = (indent + l for l in lines)

    ret.extend(lines)

    if end:
        ret.append(end)

    return ret


id_trans_table = {
    '<': '_lt',
    '>': '_gt',
    ':': '_col_',
    '*': '_star_',
    '&': '_ref_',
    '.': '_dot_'
}
id_sub_pattern = re.compile(r'(^(?![^a-zA-Z_]))|[^a-zA-Z0-9_]')


@lru_cache()
def clean_identifier(unsafe):
    """
    turns an unsafe string or object into a safe cpp identifier
    (without illegal characters, whitespace, and starting with a legal character)
    """
    if not isinstance(unsafe, str):
        if isinstance(unsafe, Iterable):
            return '_'.join(clean_identifier(i) for i in unsafe)
        unsafe = str(unsafe)

    def sub_func(match):
        return id_trans_table.get(match[0], '_')

    return id_sub_pattern.sub(sub_func, unsafe)


class ExtendedFormatter(Formatter):
    """
    Formatter with extended conversion symbols:
    * l: lowercase
    * u: uppercase
    * c: capitalized
    """

    def convert_field(self, value, conversion):
        if conversion == 'u':
            return str(value).upper()
        elif conversion == 'l':
            return str(value).lower()
        elif conversion == 'c':
            return str(value).capitalize()
        return super().convert_field(value, conversion)


Self_Replaceable = TypeVar('Self_Replaceable', bound='Replaceable')


class Replaceable:
    """
    A base class to allow for easy copying of simple types
    NOTE: this should only be used with immutable types
    """

    def __new__(cls, *args, **kwargs):
        ret = super().__new__(cls)
        ret.__args = (args, kwargs)
        return ret

    @property
    def can_replace(self):
        return True

    def replace(self: Self_Replaceable, **changes) -> Self_Replaceable:
        """
        create a new object of the type of self, using self's arguments as template
        :param kwargs: initialization arguments to change
        :return: a new instance of self's type, using a mix of self's init arguments and kwargs
        """
        if not self.can_replace:
            raise Exception(f'object of type {type(self)} is in a non-replaceable state')

        if not changes:
            return self

        args = list(self.__args[0])
        kwargs = dict(self.__args[1])
        for k, v in changes.items():
            param_index = self.__init_params.get(k, -1)
            if 0 <= param_index < len(args):
                args[param_index] = v
            else:
                kwargs[k] = v
        return type(self)(*args, **kwargs)

    def __init_subclass__(cls, **kwargs):
        # to allow kwargs->args replacement, we need a dict of all of the class's original positional arguments
        super().__init_subclass__(**kwargs)
        update_wrapper(cls.replace, cls.__init__)
        try:
            parent_params = cls.__init_params
        except AttributeError:
            parent_params = {}

        cls.__init_params = {}
        for i, p in enumerate(signature(cls.__init__).parameters.values()):
            if p.kind == Parameter.VAR_POSITIONAL:
                for n, j in parent_params.items():
                    cls.__init_params[n] = j + i - 1

            if p.kind not in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.POSITIONAL_ONLY):
                break

            cls.__init_params[p.name] = i - 1

    # all the other methods here are rudimentary but they're useful to have
    def __repr__(self):
        args = chain(
            (repr(a) for a in self.__args[0]),
            (str(k) + ' = ' + repr(v) for k, v in self.__args[1].items())
        )
        return type(self).__name__ + '(' + ', '.join(args) + ')'

    def __eq__(self, other):
        return super().__eq__(self, other) \
               or (self.can_replace and isinstance(other, type(self)) and other.can_replace and
                   (self.__args == other.__args))

    def __hash__(self):
        return hash(
            (type(self), self.__args[0], tuple(self.__args[1].items()))
        )


__all__ = [
    'Verbosity',
    'scope_lines', 'clean_source', 'clean_identifier',
    'ExtendedFormatter',
    'Replaceable'
]



from swimport.swim_run import swimport_paths_tuple


def get_path_set(paths_set=0, python_major=3, python_minor=7) -> swimport_paths_tuple:
    PYTHON_ROOT = os.path.dirname(sys.executable)
    SWIG_PATH = r'd:\dev\swigwin-3.0.12\swig.exe'
    if not paths_set:
        windows_kit_template = r'C:\Program Files (x86)\Windows Kits\10\{}\10.0.17763.0' + "\\"
        MSVC_dir = r'C:\Program Files (x86)\Microsoft Visual Studio\2017\Community\VC\Tools\MSVC\14.16.27023' + "\\"
    else:
        windows_kit_template = r'C:\Program Files (x86)\Windows Kits\10\{}\10.0.18362.0' + "\\"
        MSVC_dir = r'c:\Program Files (x86)\Microsoft Visual Studio\2019\Professional\VC\Tools\MSVC\14.26.28801' + "\\"

    np_include_path = PYTHON_ROOT + r"\Lib\site-packages\numpy\core\include\\"
    PY_INCLUDE_PATH = PYTHON_ROOT + r'\include'
    PY_LIB_PATH = PYTHON_ROOT + r'\libs\python{}{}.lib'.format(python_major, python_minor)

    windows_kit_include = windows_kit_template.format('include')
    windows_kit_lib = windows_kit_template.format('lib')

    CL_PATH = MSVC_dir + r'bin\Hostx64\x64\cl.exe'
    COMPILE_ADDITIONAL_INCLUDE_DIRS = [
        MSVC_dir + 'include',
        windows_kit_include + 'ucrt',
        windows_kit_include + 'shared',
        windows_kit_include + 'um',
        np_include_path
    ]
    COMPILE_ADDITIONAL_LIBS = [
        MSVC_dir + r'lib\x64\libcpmt.lib',
        MSVC_dir + r'lib\x64\libcmt.lib',
        MSVC_dir + r'lib\x64\oldnames.lib',
        MSVC_dir + r'lib\x64\libvcruntime.lib',
        windows_kit_lib + r'um\x64\kernel32.lib',
        windows_kit_lib + r'ucrt\x64\libucrt.lib',
        windows_kit_lib + r'um\x64\Uuid.lib'
    ]
    return swimport_paths_tuple(
        SWIG_PATH, PYTHON_ROOT, windows_kit_template, MSVC_dir, np_include_path, PY_INCLUDE_PATH,
        PY_LIB_PATH, windows_kit_include, windows_kit_lib, CL_PATH, COMPILE_ADDITIONAL_INCLUDE_DIRS,
        COMPILE_ADDITIONAL_LIBS)



from typing import Iterator, Dict, Iterable

import re

from swimport.swim import Swim
from swimport.__util__ import *
from swimport.model import Swimporting, SwimportAggregate, Typedef, CPPObject, Compileable, object_swimporting


class TypedefAggregate(Iterable[Swimporting], SwimportAggregate):
    """
    An object that acts like a swim. But accepts only typedef swimportings.
    It can then be iterated to sort the swimported typedefs topologically.
    >>> ta = TypedefAggregate()
    >>> class Mock_Swimporting: pass
    >>> def ap(d, s):
    ...     app = Mock_Swimporting()
    ...     app.object = Typedef.make(str(d), str(s))
    ...     ta(app)
    >>> ap(0, 1)
    >>> ap(2, 3)
    >>> ap(1, 2)
    >>> list((t.object.dest_type, t.object.source_type) for t in ta)
    [('2', '3'), ('1', '2'), ('0', '1')]
    """

    def __init__(self):
        self.imported_typedefs: Dict[Typedef, Swimporting] = {}

    def apply_porting(self, swimporting: Swimporting):
        """
        store a swimporting of a typedef
        :param swimporting: the swimporting object
        :return: whether the object was successfully imported (returns 0 if the typedef wal already imported)
        :raises TypeError: if swimporting is not an ObjectSwimporting[Typedef]
        """
        obj: CPPObject = getattr(swimporting, 'object', None)
        if not isinstance(obj, Typedef):
            raise TypeError('Typedef aggregate can only accept swimportings of typedefs')
        v = self.imported_typedefs.setdefault(obj, swimporting)
        if v is swimporting:
            return obj.name
        return 0

    def __iter__(self) -> Iterator[Swimporting]:
        """
        :return: all the swimportings stored in the aggregate in topological order
        """
        mapping: Dict[str, Swimporting] = {k.dest_type: v for k, v in self.imported_typedefs.items()}
        while mapping:
            _, td = mapping.popitem()
            stack = [td]
            while td.object.source_type.strip("* &") in mapping:
                td = mapping.pop(td.object.source_type.strip("* &"))
                stack.append(td)
            yield from reversed(stack)


class TypedefSourceTrigger(Typedef.Trigger):
    """A trigger for typedefs. See type_matching for an explanation on type pattern matching"""

    def __init__(self, pattern: Compileable):
        self.pattern = re.compile(pattern)

    def is_valid(self, rule: 'SwimRule', obj):
        return super().is_valid(rule, obj)\
            and self.pattern.fullmatch(obj.normalized_type)


@Typedef.set_default_trigger
class TypedefDestinationTrigger(Typedef.Trigger):
    """A trigger for typedefs"""

    def __init__(self, destination_type_pattern: Compileable):
        """
        :param destination_type_pattern: pattern the destination type (the new type) must fully match to be valuid
        """
        self.destination_type_pattern = re.compile(destination_type_pattern)

    def is_valid(self, rule, obj: Typedef):
        return super().is_valid(rule, obj) \
               and self.destination_type_pattern.fullmatch(obj.dest_type)


@Typedef.set_default_behaviour
class TypedefBehaviour(Typedef.Behaviour):
    DEFAULT_MAPS = ('{} &OUTPUT', '{} *OUTPUT',
                    '{} &INPUT', '{} *INPUT',
                    '{} &INOUT', '{} *INOUT',
                    '{} const *INPUT', '{} const &INPUT',
                    '({}* ARGOUTVIEW_ARRAY1, size_t* DIM1)')
    """A behaviour for typedefs"""

    def __init__(self, maps=DEFAULT_MAPS, inline=False, no_declare=False):
        """
        :param maps: the typemaps to apply to the new type. default is both OUTPUT and ARR_OUT
        :param inline: whether to inline the typedef's declaration
        :param no_declare: whether to forgo declaring the typedef entirely (leaving only the maps)
        """
        self.maps = list(maps)
        self.inline = inline
        self.no_declare = no_declare

    @object_swimporting
    def wrap(self, rule, obj: Typedef, swim):
        mappings = []
        mappings.extend(
            '%apply ' + m.format(obj.source_type) + ' {' + m.format(obj.dest_type) + '};'
            for m in self.maps
        )
        if self.inline:
            body_scope = Swim.add_inline
        elif self.no_declare:
            body_scope = None
        else:
            body_scope = Swim.add_raw

        swim.add_comment(f'typedef {obj.source_type} {obj.dest_type};')
        swim.add_comment(f'rule used: {rule}', verbosity_level=Verbosity.debug)
        swim.add_raw(mappings)
        if body_scope:
            body_scope(swim, f'typedef {obj.source_type} {obj.dest_type};')
        swim.add_nl()

        if swim.types.get(obj.dest_type):
            raise Exception('the destination type of the typedef ('+obj.dest_type+') was previously imported.')
        src_porting = swim.types.get(obj.source_type)
        if src_porting:
            swim.types[obj.dest_type] = src_porting



import swimport.pools.types
import swimport.pools.derived_types
import swimport.pools.complex_types
from swimport.pools.pools import pools, Pool, syspools

del types, derived_types, complex_types

pools.auto_build_doc = True
pools.rebuild_doc()


from pathlib import Path
import os
import sys
import itertools as it
import subprocess
import warnings
import ntpath
import glob
from collections import namedtuple

"""
An open letter to the c++ compiler:
Dear c++ compiler,
I get that performance is paramount, I get it. turning this:
return add(x,y);

into

return x+y;
is all well and good, but turning this:
auto x = A()
B;
return x;

into:
B;
return A();

is terrible, and leads to the worst bugs imaginable. So thank you c++ compiler. Thank you for thinking that time travel
is okay, you are the sole reason that the keyword `volatile` appears almost as much as `const`.

yours faithfully- now and until rust has normal overloads,
Ben
"""


fields = ('SWIG_PATH', 'PYTHON_ROOT', 'windows_kit_template', 'MSVC_dir', 'np_include_path', 'PY_INCLUDE_PATH',
          'PY_LIB_PATH', 'windows_kit_include', 'windows_kit_lib', 'CL_PATH', 'COMPILE_ADDITIONAL_INCLUDE_DIRS',
          'COMPILE_ADDITIONAL_LIBS')

swimport_paths_tuple = namedtuple('swimport_paths', fields)


def swim_run(
        dirname: Path,
        swimport_paths: swimport_paths_tuple,
        optimization,
        print_out=True, module_name=..., is_cpp=...):
    original_cwd = os.getcwd()
    os.chdir(str(dirname.absolute()))

    if module_name is ...:
        files = glob.glob(str(dirname / '*.m'))
        if files:
            head, tail = ntpath.split(files[0])
            parts = tail.split('.')
            module_name = parts[0]
            if is_cpp is ...:
                is_cpp = 'c' not in parts
        else:
            module_name = 'example'
    if is_cpp is ...:
        is_cpp = True

    cxx_ext = '.cxx' if is_cpp else '.c'
    src_ext = '.cpp' if is_cpp else '.c'

    main_path = dirname / 'main.py'
    inter_path = dirname / Path(module_name+'.i')
    cxx_path = dirname / (module_name+'_wrap'+cxx_ext)
    usage_path = dirname / 'usage.py'

    if main_path.exists():
        if print_out:
            print('\t' + 'main.py', end='...', flush=True)
        subprocess.run([sys.executable, 'main.py', module_name],
                       stdout=subprocess.PIPE, check=True)
        if print_out:
            print('done!')
    elif print_out:
        print('\tmain.py missing')

    if inter_path.exists():
        if print_out:
            print('\t' + module_name+'.i', end='...', flush=True)
        popenargs = [swimport_paths.SWIG_PATH]
        if is_cpp:
            popenargs.append('-c++')
        popenargs.extend(['-python', '-py3'])
        # popenargs.append('-debug-tmsearch')
        popenargs.extend(['-outdir', str(dirname), str(inter_path)])

        subprocess.run(popenargs, check=True)
        if print_out:
            print('done!')
    elif print_out:
        print('\t'+module_name+'.i missing')

    if cxx_path.exists():
        if print_out:
            print('\t' + str(cxx_path), end='...', flush=True)
        tmpdir = dirname / 'temp'
        src_path = dirname / Path('src'+src_ext)

        tmpdir.mkdir(exist_ok=True, parents=True)
        # todo compile with warnings?

        is_64bits = sys.maxsize > 2 ** 32
        pyd_suffix = f'.cp{sys.version_info[0]}{sys.version_info[1]}-{"win_amd64" if is_64bits else "win32"}.pyd'

        proc = subprocess.run([
            swimport_paths.CL_PATH, '/nologo', '/LD', '/EHsc', '/utf-8', optimization,
            '/Tp', str(cxx_path),
            *(('/Tp', str(src_path)) if src_path.exists() else ()),
            '/Fo:' + str(tmpdir) + '\\',
            '/I', swimport_paths.PY_INCLUDE_PATH,
            '/I', str(dirname),
            *it.chain.from_iterable(('/I', i) for i in swimport_paths.COMPILE_ADDITIONAL_INCLUDE_DIRS),
            '/link',
            '/LIBPATH', swimport_paths.PY_LIB_PATH,
            *it.chain.from_iterable(('/LIBPATH', l) for l in swimport_paths.COMPILE_ADDITIONAL_LIBS),
            '/IMPLIB:' + str(tmpdir / Path(module_name+'.lib')),
            '/OUT:' + str(dirname / Path('_'+module_name+pyd_suffix))
        ], stdout=subprocess.PIPE, text=True)
        if proc.returncode != 0:
            print(proc.stdout)
            raise Exception(f'cl returned {proc.returncode}')
        if print_out:
            print('done!')
    elif print_out:
        print('\t'+str(cxx_path)+' missing')

    if usage_path.exists():
        if print_out:
            print('\t' + 'usage.py', end='...', flush=True)
        subprocess.run([sys.executable, 'usage.py'],
                       stdout=None, stderr=None, check=True
                       )
        if print_out:
            print('done!')
    elif print_out:
        print('\tusage.py missing')

    os.chdir(original_cwd)


def simple_passes_filter(path: Path):
    head, tail = ntpath.split(path)
    return not tail.startswith('_')


def swim_run_subfolders(root_dir, optimization, passes_filter, **kwargs):
    if optimization != '/O2':
        warnings.warn(f'warning: optimization should be set to maximum (/O2 instead of {optimization}),'
                      'to ensure completeness',
                      stacklevel=10)
    for item in root_dir.iterdir():
        if not item.is_dir() or not passes_filter(item):
            continue
        print(str(item.name) + ':')
        swim_run(item, optimization=optimization, **kwargs)



from typing import List

from swimport.pools.pools import pools, syspools
from swimport.pools.types import TypeSwimportingPool
from swimport.typeswim import TypeSwimporting, FunctionBody


@syspools.add(TypeSwimportingPool)
def empty_tuple(**type_args):
    return TypeSwimporting('std::tuple<>', 'Tuple[]',
                           to_py="""
                           return PyTuple_New(0);
                           """,
                           to_cpp="""
                           if (!PySequence_Check(input)) {
                                ok = 0;
                                PyErr_SetString(PyExc_TypeError, "expected sequence");
                                return;
                            }
                            if (PySequence_Size(input) != 0) {
                                ok = 0;
                                PyErr_SetString(PyExc_ValueError, "expected exactly 0 items in sequence");
                                return;
                            }
                            ret* = std::make_tuple<>();""",
                           to_cpp_check="""
                           if (!PySequence_Check(input)) {
                                return false;
                            }
                            if (PySequence_Size(input) != 0) {
                                return false;
                            }
                            return true;"""
                           , **type_args)


@pools.add(TypeSwimportingPool, name='tuple')
def tuple_(*types, cpp_name='std::tuple<?>', pair=..., swim, **type_args):
    """
    typemaps for using std::tuple for a set of types
    :param types: the types to handle
    :param cpp_name: the name of the cpp type to map for. ? will be replaced with the type names
    :param pair: whether to also copy the maps for the std::pair type. By default will only apply to pair if the cpp
        name is std::tuple and there are exactly 2 types.
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    if not types:
        return empty_tuple.get(**type_args)

    sub_portings: List[TypeSwimporting] = []
    types = list(types)
    for i, t in enumerate(types):
        t, sp = swim.get_porting(t)
        types[i] = t
        sub_portings.append(sp)

    cpp_name = cpp_name.replace('?', ', '.join(types))

    to_py_funcs = [sp.to_py_func for sp in sub_portings]
    to_cpp_funcs = [sp.to_cpp_func for sp in sub_portings]
    to_cpp_check_funcs = [sp.to_cpp_check_func for sp in sub_portings]
    to_cpp_post_funcs = [sp.to_cpp_post_func for sp in sub_portings]

    if all(to_py_funcs):
        body = ['ok = 1;',
                f'auto ret = PyTuple_New({len(types)}); if (!ret) {{ok = false; return ret;}};']
        for i, func in enumerate(to_py_funcs):
            body.append(f"""
            {{
                PyObject* _pyval = {func.function_name}(std::get<{i}>(input), ok, PACKED_UD({i}));
                if (!ok || PyErr_Occurred()) {{ return ret; }}
                PyTuple_SET_ITEM(ret, {i}, _pyval);
            }}""")
        body.append("""
            return ret;
            """)

        to_py_func = FunctionBody.combine(*to_py_funcs)(body)
    else:
        to_py_func = None

    if all(to_cpp_funcs):
        body = ['ok = 1;', 'PyObject* _p = nullptr;']
        for i, func in enumerate(to_cpp_funcs):
            body.append(f'{func.owner.cpp_name} _{i};')
        default_tuple = 'std::make_tuple(' + ', '.join(f'_{i}' for i in range(len(to_cpp_funcs))) + ')'
        body.append(
            f"""
            if (!PySequence_Check(input)) {{
                ok = 0;
                PyErr_SetString(PyExc_TypeError, "expected sequence");
                return {default_tuple};
            }}
            if (PySequence_Size(input) != {len(types)}) {{
                ok = 0;
                PyErr_SetString(PyExc_ValueError, "expected exactly {len(types)} items in sequence");
                return {default_tuple};
            }}""")
        for i, func in enumerate(to_cpp_funcs):
            body.append(
                f"""
                _p = PySequence_ITEM(input, {i});
                if (!_p) return {default_tuple};
                _{i} = {func.function_name}(_p, ok, PACKED_UD({i}));
                Py_XDECREF(_p);
                if (!ok || PyErr_Occurred()) return {default_tuple};""")
        body.append(f'return {default_tuple};')

        to_cpp_func = FunctionBody.combine(*to_cpp_funcs)(body)
    else:
        to_cpp_func = None

    if all(to_cpp_check_funcs):
        body = []
        for i, func in enumerate(to_cpp_check_funcs):
            body.append(f'PyObject* _p{i} = nullptr;')
        body.append(f'if (!PyArg_ParseTuple(input, "{"O" * len(types)}", ' + ', '.join(
            '&_p' + str(i) for i in range(len(types))) + f')) return 0;')
        for i, func in enumerate(to_cpp_check_funcs):
            body.append(
                f'if(!{func.function_name}(_p{i}, PACKED_UD({i}))) {{return 0;}}')
        body.append(f'return 1;')

        to_cpp_check_func = FunctionBody.combine(*to_cpp_check_funcs)(body)
    else:
        to_cpp_check_func = None

    if any(to_cpp_post_funcs):
        body = ['ok = 1;']
        for i, func in enumerate(to_cpp_post_funcs):
            if not func:
                continue
            body.append(f'{func.function_name}(std::get<{i}>(input), ok, PACKED_UD({i}));')
        to_cpp_post_func = FunctionBody.combine(*to_cpp_post_funcs)(body)
    else:
        to_cpp_post_func = None

    if pair is ...:
        pair = (cpp_name.startswith('std::tuple<') or cpp_name.startswith('tuple<')) \
               and len(types) == 2
    if pair:
        aliases = ('std::pair<' + ', '.join(types) + '>',)
    else:
        aliases = ()

    return TypeSwimporting(cpp_name, 'Tuple[' + ', '.join(sp.py_name for sp in sub_portings) + ']',
                           to_py=to_py_func,
                           to_cpp=to_cpp_func,
                           to_cpp_check=to_cpp_check_func,
                           to_cpp_post=to_cpp_post_func,
                           aliases=aliases, **type_args)



from typing import List

from itertools import chain
from warnings import warn

from swimport.pools.pools import pools
from swimport.pools.types import TypeSwimportingPool
from swimport.typeswim import TypeSwimporting, FunctionBody


@pools.add(TypeSwimportingPool, name='callable')
def callable_(ret_type, *arg_types, cpp_name='std::function<?>', surrogate_type_name='py_callable<?>',
              on_ret_post='raise', swim,
              **type_args):
    """
    typemaps for using std::function for a set of types
    :param ret_type: the return type of the function, or void (or None)
    :param arg_types: the types of the arguments
    :param cpp_name: the name of the cpp type to map for. ? will be replaced with the type names
    :param surrogate_type_name: the type name of the surrogate object to adapt between a python callable and
        a c++ object.
    :param on_ret_post: What to do if the return type has a to_cpp_post map, 'raise', 'warn', or None.
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    arg_portings: List[TypeSwimporting] = []
    if ret_type == 'void':
        ret_type = None

    if ret_type:
        ret_type, ret_porting = swim.get_porting(ret_type)
        if not ret_porting.to_cpp_func:
            raise ValueError('the return type ' + ret_type + ' does not have a to_cpp function')
        if ret_porting.to_cpp_post_func:
            msg = f"the return type {ret_type} has a to_cpp_post function, that the object will not be able to"\
                    f" dispose of, memory leaks possible"
            if on_ret_post == 'warn':
                warn(msg)
            elif on_ret_post is None:
                pass
            else:
                raise ValueError(msg)
    else:
        ret_type = 'void'
        ret_porting = None

    arg_types = list(arg_types)
    for i, t in enumerate(arg_types):
        t, sp = swim.get_porting(t)
        if not sp.to_py_func:
            raise ValueError('the type ' + t + ' does not have a to_py function')
        arg_portings.append(sp)
        arg_types[i] = t

    cpp_name = cpp_name.replace('?', ret_type + '(' + ', '.join(arg_types) + ')')
    surrogate_type_name = surrogate_type_name.replace('?', ', '.join(chain((ret_type,), arg_types)))

    swim(pools.include("<tuple>"))

    to_cpp_body = [
        f"""
        if (!PyCallable_Check(input)){{
            PyErr_SetString(PyExc_TypeError, "expected a callable object");
            return {{}};
        }}
        auto to_py_conv_pack = std::make_tuple({','.join(
            "(py_call::converter_to_py<" + t + ">)PACKED_UD(" + str(i) + ")" for i, t in enumerate(arg_types))});
        """
    ]
    if ret_porting:
        to_cpp_body.append(
            f"auto to_cpp_conv = (py_call::converter_to_cpp<{ret_type}>)PACKED_UD({len(arg_types)});"
        )
    else:
        to_cpp_body.append(
            f"void* to_cpp_conv = nullptr;"
        )
    to_cpp_body.append(
        f"""
        {surrogate_type_name} surrogate(to_cpp_conv, to_py_conv_pack, input);
        return ({cpp_name})std::bind( std::mem_fn(&{surrogate_type_name}::call) , {", ".join(
            chain(("surrogate",), ('std::placeholders::_' + str(i) for i, _ in enumerate(arg_types, 1))))});
        """
    )
    if not (arg_portings or ret_porting):
        userdata = ''
        frags = ()
    else:
        userdata = [f'void* volatile ud[{len(arg_portings) + bool(ret_porting)}] = {{}};']
        frags = []
        for i, sp in enumerate(arg_portings):
            fb = sp.to_py
            t = sp.cpp_name
            userdata.append(f"""
            {{
                ud[{i}] = (py_call::converter_to_py<{t}>)[]({t} const & input,int& ok){{ 
                    void* volatile userdata=nullptr;  // userdata for type {t}
                    {{
                    {fb.user_data}
                    }}
                    return {fb.function_name}(input, ok, userdata);
                }};
            }}
            """)
            frags.extend(fb.frags)
        if ret_porting:
            fb = ret_porting.to_cpp
            t = ret_type
            userdata.append(f"""
            {{
                ud[{len(arg_portings)}] = (py_call::converter_to_cpp<{t}>)[](PyObject* input,int& ok){{ 
                    void* volatile userdata=nullptr;  // userdata for type {t}
                    {{
                    {fb.user_data}
                    }}
                    return {fb.function_name}(input, ok, userdata);
                }};
            }}
            """)
            frags.extend(fb.frags)

        userdata.append('userdata = (void* volatile)ud;')
        userdata = '\n'.join(userdata)

    to_cpp = FunctionBody(to_cpp_body, user_data=userdata, fragments=frags)

    return TypeSwimporting(cpp_name, py_name='Callable',
                           to_cpp=to_cpp,
                           to_cpp_check='return PyCallable_Check(input);',
                           **type_args)


from swimport.pools.pools import pools
from swimport.pools.types import TypeSwimportingPool
from swimport.typeswim import TypeSwimporting, FunctionBody


@pools.add(TypeSwimportingPool, name='iter')
def iter_(inner_type, *, cpp_name='std::vector<?>', input_=..., insert_func=..., reserve=..., output=iter, swim,
          **type_args):
    """
    typemaps for iterable type
    :param inner_type: the type to wrap in iterable input
    :param cpp_name: the name of the type to use in maps.
    :param input_: whether to include input typemaps for the type. ... is to only include if available.
    :param insert_func: the name of the function to insert values into the cpp type. Default is to guess based on
        cpp_name. Only used if input is true.
    :param reserve: whether to call reserve on the C++ structure using an input's __length_hint__. Default is
        to only reserve for std::vector. Only used if input is true.
    :param output: whether to include output typemaps for the type. False, for nothing.
        Default is to use an iterable adapter.
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    type_args.setdefault('in_iterable_types', ())
    type_args.setdefault('out_iterable_types', ())
    type_args.setdefault('inout_maps', False)

    inner_type, inner_porting = swim.get_porting(inner_type)

    cpp_name = cpp_name.replace('?', inner_type)

    if not inner_porting.to_cpp_func:
        if input_ is True:
            raise Exception(f'type {inner_type} has no to_cpp function')
        elif input_ is ...:
            input_ = False
    elif input_ is ...:
        input_ = True

    if not inner_porting.to_py_func:
        if output is True:
            raise Exception(f'type {inner_type} has no to_py function')
        elif output is ...:
            output = False
    elif output is ...:
        output = True

    if output is iter:
        iterable_type = 'py_wrapper_iterable<' + inner_type + ', ' + cpp_name + '>'
        converter_name = 'py_iter::converter_to_py<?>'.replace('?', inner_type)
        to_py = FunctionBody(f"""
                    using rtype = {iterable_type};
                    rtype* a = PyObject_New(rtype, &rtype::typeObject);
                    new(a) rtype(({converter_name})userdata, input);
                    return (PyObject*)a;
                   """,
                             user_data=f"""userdata = ({converter_name})([]({inner_type} const & v, int& ok)
                                            {{
                                                void* volatile userdata=nullptr;
                                                {inner_porting.to_py_func.user_data}
                                                return {inner_porting.to_py_func.function_name}(v, ok, userdata);
                                            }});
                                           """,
                             fragments=inner_porting.to_py_func.frags)
    elif output is list:
        to_py = FunctionBody.combine(inner_porting.to_py_func)(f"""
            ok = 1;
            auto ret = PyList_New(input.size());
            if (!ret){{
                ok = 0;
                return nullptr; 
            }}
            Py_ssize_t  i = 0;
            for (auto e: input){{
                PyObject* to_add = {inner_porting.to_py_func.function_name}(e, ok, PACKED_UD(0));
                if (!ok || PyErr_Occurred()){{
                    Py_CLEAR(ret);
                    return nullptr;
                }}             
                PyList_SET_ITEM(ret, i, to_add); 
                if (!ok || PyErr_Occurred()){{
                    Py_XDECREF(ret);
                    return nullptr;
                }}
                i++;
            }}
            return ret;
        """)
    elif output is tuple:
        to_py = FunctionBody.combine(inner_porting.to_py_func)(f"""
                ok = 1;
                auto ret = PyTuple_New(input.size());
                if (!ret){{
                    ok = 0;
                    return nullptr; 
                }}
                Py_ssize_t  i = 0;
                for (auto e: input){{
                    PyObject* to_add = {inner_porting.to_py_func.function_name}(e, ok, PACKED_UD(0));
                    if (!ok || PyErr_Occurred()){{
                        Py_CLEAR(ret);
                        return nullptr;
                    }}                
                    PyTuple_SET_ITEM(ret, i, to_add); 
                    i++;
                }}
                return ret;
                """)
    elif output is set:
        to_py = FunctionBody.combine(inner_porting.to_py_func)(f"""
                ok = 1;
                auto ret = PySet_New(nullptr);
                if (!ret){{
                    ok = 0;
                    return nullptr; 
                }}
                for (auto e: input){{
                    auto to_add = {inner_porting.to_py_func.function_name}(e, ok, PACKED_UD(0));
                    if (!ok || PyErr_Occurred()){{
                        SWIM_THROW_UNSPECIFIED;
                    }}                
                    if (PySet_Add(ret, to_add)){{
                        Py_DECREF(to_add);
                        SWIM_THROW_UNSPECIFIED;
                    }} 
                }}
                return ret;
                """)
    elif output is frozenset:
        to_py = FunctionBody.combine(inner_porting.to_py_func)(f"""
                ok = 1;
                auto ret = PyFrozenSet_New(nullptr);
                if (!ret){{
                    ok = 0;
                    return nullptr; 
                }}
                for (auto e: input){{
                    auto to_add = {inner_porting.to_py_func.function_name}(e, ok, PACKED_UD(0));
                    if (!ok || PyErr_Occurred()){{
                        SWIM_THROW_UNSPECIFIED;
                    }}                
                    if (PySet_Add(ret, to_add)){{
                        Py_DECREF(to_add);
                        SWIM_THROW_UNSPECIFIED;
                    }} 
                }}
                return ret;
                """)
    elif not output:
        to_py = None
    else:
        raise TypeError('bad type for output')

    if input_:
        if insert_func is ...:
            if any((cpp_name.startswith(prefix) or cpp_name.startswith('std::' + prefix))
                   for prefix in ('vector<', 'deque<', 'list<', 'forward_list<')):
                insert_func = 'push_back'
            elif any((cpp_name.startswith(prefix) or cpp_name.startswith('std::' + prefix))
                     for prefix in ('unordered_set<', 'set<')):
                insert_func = 'insert'
            else:
                raise Exception('could not guess insert_func for type ' + cpp_name)

        if reserve is ...:
            reserve = any(
                cpp_name.startswith(pref) for pref in ('std::vector<', 'vector<')
            )

        to_cpp = FunctionBody.combine(inner_porting.to_cpp_func)(
            f"""
            ok = 1;
            auto iter = PyObject_GetIter(input);
            if (!iter){{
                SWIM_RAISE_TYPE("expected an iterable, not ", input);
            }}
            {cpp_name} ret;
            """
            +
            ("""
                auto size = PyObject_LengthHint(input, -2);
                if (size >= 0)
                    ret.reserve(size);
                else if (size != -2){
                    SWIM_RAISE_EXISTING;
                }
            """ if reserve else "")
            +
            f"""                
            PyObject* current;
            while(current = PyIter_Next(iter)){{
                {inner_type} cpp_curr = {inner_porting.to_cpp_func.function_name}(current, ok, PACKED_UD(0));
                Py_DECREF(current);
                if (!ok || PyErr_Occurred()){{
                    PyErr_SetString(PyExc_TypeError, "could not convert object in iterable"); 
                    SWIM_RAISE_UNSPECIFIED;
                }}
                ret.{insert_func}(cpp_curr);
            }}
            Py_DECREF(iter);
            return ret;
            """)

        to_cpp_check = """
                    auto iter = PyObject_GetIter(input);
                    auto ret = (iter != nullptr);
                    Py_XDECREF(iter);
                    PyErr_Clear();
                    return ret;
                    """

        if inner_porting.to_cpp_post_func:
            to_cpp_post = FunctionBody.combine(inner_porting.to_cpp_post_func)(f"""
            for (auto e: input){{
                {inner_porting.to_cpp_post_func.function_name}(e, ok, PACKED_UD(0));
                if (!ok || PyErr_Occurred())
                    return;
            }}
            """)
        else:
            to_cpp_post = None

    else:
        to_cpp = None
        to_cpp_check = None
        to_cpp_post = None

    return TypeSwimporting(cpp_name, 'Iterable[' + inner_porting.py_name + ']',
                           to_py=to_py,
                           to_cpp=to_cpp,
                           to_cpp_check=to_cpp_check,
                           to_cpp_post=to_cpp_post,
                           **type_args)


set_ = pools.add(..., name='set', doc='as `iter`, but specialized for sets')(
    iter_(cpp_name='std::unordered_set<?>', output=set))

frozenset_ = pools.add(..., name='frozenset', doc='as `iter`, but specialized for frozensets')(
    iter_(cpp_name='std::unordered_set<?>', output=frozenset))

list_ = pools.add(..., name='list', doc='as `iter`, but specialized for lists')(
    iter_(cpp_name='std::vector<?>', output=list))




from setuptools import setup

from swimport_data.__data__ import (
    __name__,
    __author__,
    __author_email__,
    __license__,
    __url__,
    __version__,
)

setup(
    name=__name__,
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    license=__license__,
    url=__url__,
    packages=['swimport', 'swimport.model', 'swimport.pools', 'swimport_data', 'swimport.pools.derived_types'],
    install_requires=['CppHeaderParser>=2', ],
    python_requires='>=3.6.0',
    include_package_data=True,
    data_files=[
        ('', ['README.md', 'CHANGELOG.md']),
    ],
    extras_require={
        'memory testing': ['psutil']
    },
)
